/*
  Copyright (C) 2010 Alexander Vorobiev.  All rights reserved.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  See file LICENSE.txt for further informations on licensing terms.
*/



#include "RailwayHardware.h"
#include "WProgram.h"


// ------------------------------------------- TRailwayHardware ----------------------------------------------------

//  -   
#define EXTENDER_IOS		9
//   
#define CROSS_COUNT			3


TRailwayHardware::TRailwayHardware(void)
: m_Io(EXTENDER_IOS), m_CrossSens(CROSS_COUNT) {
}


//    setup
void TRailwayHardware::Setup(void) {
	m_Io.Setup();
	m_LightsCtrl.Setup(&m_Io);
	m_CrossSens.Setup(&m_Io);
	m_SwCtrl.Setup();
	m_TrainCtrl.Setup();
	m_BtnCtrl.Setup(&m_Io);
}


//     loop
void TRailwayHardware::LoopCall(void) {
	m_CrossSens.LoopCall();
	m_SwCtrl.LoopCall();
	m_TrainCtrl.LoopCall();
	m_BtnCtrl.LoopCall();
	m_Io.ReadWrite();
}


// -------------------------------------- TRailwayLightsControl ----------------------------------------------------


void TRailwayLightsControl::Setup(TInOutExtenderS* io) {
	m_pIo = io;
}

//  
void TRailwayLightsControl::Red(uint16_t lt) {
	m_pIo->m_Out &= (~lt);
}

//  
void TRailwayLightsControl::Green(uint16_t lt) {
	m_pIo->m_Out |= lt;
}


// -------------------------------------- TRailwayCrossSensors -----------------------------------------------------

//   (  )     
#define	CROSS_CTRL		0x0100
//   / 
#define CROSS_PERIOD	50
//   ( )
#define CROSS_DELTA		10
//        ""
#define CROSS_ON_CNT	3
//        " "
#define CROSS_OFF_CNT	10

TRailwayCrossSensors::TRailwayCrossSensors(uint8_t count) {
	m_Count = count;
	m_PrevSwitchTime = 0;
}


void TRailwayCrossSensors::Setup(TInOutExtenderS* io) {
	m_pIo = io;
	analogReference(INTERNAL);
	uint8_t i;
	for(i=0;i<m_Count;i++) {
		m_CrossSignal[i] = 0;
		m_CrossState[i] = 0;
		m_CrossCounter[i] = 0;
		m_CrossOn[i] = 0;
		m_CrossOff[i] = 0;
	}
}


//  
void TRailwayCrossSensors::LoopCall(void) {
	//     / o
	unsigned long tm = millis();
	if( tm - m_PrevSwitchTime < CROSS_PERIOD ) return; //   

	// 
	m_PrevSwitchTime = tm;

	//        
	uint8_t i;
	int value,delta;
	for(i=0;i<m_Count;i++) {
		//  
		value = analogRead(i);
		//    
		delta = value - m_AnalogValue[i];
/*
		if(i==2) {
			Serial.print( value ); Serial.print(" ");
			Serial.print( m_AnalogValue[i] ); Serial.print(" ");
			Serial.print( delta );
			Serial.println("");
		}
*/
		//   
		m_CrossSignal[i] = (abs(delta) < CROSS_DELTA);
		//     
		m_AnalogValue[i] = value;

		//    
		m_CrossOn[i] = 0;
		m_CrossOff[i] = 0;

		if( m_CrossState[i] ) {
			//    ""
			if( m_CrossSignal[i] ) {
				//   - 
				m_CrossCounter[i] = 0;
			} else {
				//   -  
				m_CrossCounter[i]++;
				if( m_CrossCounter[i] >= CROSS_OFF_CNT ) {
					//    
					m_CrossCounter[i] = 0;
					m_CrossState[i] = 0;
					m_CrossOff[i] = 1;
				}
			}
		} else {
			//    " "
			if( m_CrossSignal[i] ) {
				//   - 
				m_CrossCounter[i]++;
				if( m_CrossCounter[i] >= CROSS_ON_CNT ) {
					//    
					m_CrossCounter[i] = 0;
					m_CrossState[i] = 1;
					m_CrossOn[i] = 1;
				}
			} else {
				//   -  
				m_CrossCounter[i] = 0;
			}
		}
	}



	//      
	if( m_pIo->m_Out & CROSS_CTRL ) m_pIo->m_Out &= ~CROSS_CTRL;
	else m_pIo->m_Out |= CROSS_CTRL;
}


// --------------------------------------- TRailwaySwitchControl ---------------------------------------------------
//  

//   () 
#define RWSWITCH_SMOTOR		2
//    ()
#define RWSWITCH_PERIOD		200
//    
#define RWSWITCH_STEPS		22
//    ""
#define RWSWITCH_LEFT		BACKWARD
//    ""
#define RWSWITCH_RIGHT		FORWARD

TRailwaySwitchControl::TRailwaySwitchControl(void)
: m_SMotor(100,RWSWITCH_SMOTOR) { //     -    
}


//    setup
void TRailwaySwitchControl::Setup(void) {
	m_SMotor.setSpeed( 10 ); //    
	m_Dir = 0;
	m_PrevTm = 0;
}

//  
void TRailwaySwitchControl::LoopCall(void) {
	if( ! m_Dir ) return; //   
	// - 
	//     
	unsigned long tm = millis();
	if( tm - m_PrevTm < RWSWITCH_PERIOD ) return; //  
	// 
	m_PrevTm = tm;
	m_SMotor.onestep( m_Dir, DOUBLE );
	m_Step--;
	//    
	if( ! m_Step ) {
		// 
		m_Dir = 0;
		m_SMotor.release();
	}
}


//  
void TRailwaySwitchControl::Left(void) {
	m_Step = RWSWITCH_STEPS;
	m_Dir = RWSWITCH_LEFT;
}

//  
void TRailwaySwitchControl::Right(void) {
	m_Step = RWSWITCH_STEPS;
	m_Dir = RWSWITCH_RIGHT;
}



// --------------------------------------- TRailwayTrainControl ----------------------------------------------------

//   
#define TRAIN_MOTOR		1

//  
#define	TRAIN_SPEED_COUNT	7
static uint8_t	c_TrainSpeedValue[TRAIN_SPEED_COUNT] = { 0, 125, 150, 175, 200, 225, 255 };

//    ()
#define TRAIN_SPEED_PERIOD	1000


TRailwayTrainControl::TRailwayTrainControl(void)
: m_Motor(TRAIN_MOTOR) {
}

//    setup
void TRailwayTrainControl::Setup(void) {
	m_Motor.setSpeed( 0 );
	m_Motor.run( FORWARD );
	m_Speed = 0;
	m_TgSpeed = 0;
	m_PrevTm = 0;
}


//  
void TRailwayTrainControl::LoopCall(void) {
	if( m_TgSpeed == m_Speed ) return; //    
	//    
	//  
	unsigned long tm = millis();
	if( tm - m_PrevTm < TRAIN_SPEED_PERIOD ) return; //   
	// 
	m_PrevTm = tm;
	//  
	int8_t nextSpeed = (m_TgSpeed > m_Speed) ? m_Speed + 1 : m_Speed - 1;
	//    
	if( m_Speed == 0 ) {
		if( nextSpeed > 0 ) m_Motor.run( FORWARD );
		if( nextSpeed < 0 ) m_Motor.run( BACKWARD );
	}
	//  
	m_Speed = nextSpeed;
	m_Motor.setSpeed( c_TrainSpeedValue[abs(m_Speed)] );
	//     
	//     -  
	if( m_Speed == m_TgSpeed && m_Speed == 0 ) m_Motor.run( RELEASE );
}


//  
void TRailwayTrainControl::Run(int8_t speed) {
	if( speed >= TRAIN_SPEED_COUNT ) speed = TRAIN_SPEED_COUNT-1;
	if( speed <= -TRAIN_SPEED_COUNT ) speed = -TRAIN_SPEED_COUNT+1;
	m_TgSpeed = speed;
}


// --------------------------------------- TRailwayButtonsControl --------------------------------------------------

//  
#define	BUTTONS_COUNT		3
//  
static uint16_t	c_ButtonMask[BUTTONS_COUNT] = { 0x0001, 0x0002, 0x0004 };
//  
#define BUTTONS_PERIOD		100
//      
#define BUTTONS_LOOPCNT		3

TRailwayButtonsControl::TRailwayButtonsControl(void) {
}

void TRailwayButtonsControl::Setup(TInOutExtenderS* io) {
	m_pIo = io;
	uint8_t i;
	for(i=0;i<BUTTONS_COUNT;i++) {
		m_On[i] = 0;
		m_Counter[i] = 0;
	}
}

//  
void TRailwayButtonsControl::LoopCall(void) {
	//    
	unsigned long tm = millis();
	if( tm - m_PrevTm < BUTTONS_PERIOD ) return; //   
	// 
	m_PrevTm = tm;
	//  
	uint8_t i;
	for(i=0;i<BUTTONS_COUNT;i++) {
		if( m_On[i] ) {
			//    ""
			if( m_pIo->m_In & c_ButtonMask[i] ) {
				//   
				m_Counter[i]++;
				//   
				if( m_Counter[i] >= BUTTONS_LOOPCNT ) {
					// 
					m_On[i] = 0;
					m_Counter[i] = 0;
				}
			} else {
				//  
				m_Counter[i] = 0;
			}
		} else {
			//    " "
			if( m_pIo->m_In & c_ButtonMask[i] ) {
				//   
				m_Counter[i] = 0;
			} else {
				//  
				m_Counter[i]++;
				//   
				if( m_Counter[i] >= BUTTONS_LOOPCNT ) {
					// 
					m_On[i] = 1;
					m_Counter[i] = 0;
				}
			}
		}
	}
}